import datetime
import json
import mimetypes
import os
from http import HTTPStatus

from jinja2 import Environment, FileSystemLoader

from inspira.constants import APPLICATION_JSON, NOT_FOUND, TEXT_HTML, TEXT_PLAIN, UTF8
from inspira.logging import log
from inspira.requests import RequestContext


class HttpResponse:
    def __init__(
        self,
        content=None,
        status_code=HTTPStatus.OK,
        content_type=TEXT_PLAIN,
        headers=None,
    ):
        self.content = content
        self.status_code = status_code
        self.content_type = content_type
        self.headers = dict(headers or {})

    def set_cookie(
        self,
        key,
        value,
        max_age=None,
        expires=None,
        path="/",
        domain=None,
        secure=False,
        httponly=False,
        samesite=None,
    ):
        cookie_str = f"{key}={value}"

        if max_age is not None:
            cookie_str += f"; Max-Age={max_age}"
        elif expires is not None and isinstance(expires, (int, float)):
            expires_dt = datetime.datetime.utcfromtimestamp(expires)
            expires_str = expires_dt.strftime("%a, %d %b %Y %H:%M:%S GMT")
            cookie_str += f"; Expires={expires_str}"

        if path:
            cookie_str += f"; Path={path}"
        if domain:
            cookie_str += f"; Domain={domain}"
        if secure:
            cookie_str += "; Secure"
        if httponly:
            cookie_str += "; HttpOnly"
        if samesite:
            cookie_str += f"; SameSite={samesite}"

        # Append the cookie to the headers dictionary
        self.headers.setdefault("set-cookie", []).append(cookie_str)

    async def __call__(self, scope, receive, send):
        headers = await self.encoded_headers()

        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": headers,
            }
        )

        body = await self.serialize_content()

        await send(
            {
                "type": "http.response.body",
                "body": body,
                "more_body": False,
            }
        )

    async def encoded_headers(self):
        request_headers = RequestContext().get_request().get_request_headers()

        headers = [(b"content-type", self.content_type.encode(UTF8))]

        headers.extend(request_headers)

        for key, value_list in self.headers.items():
            headers.extend(self.encode_header(key, value_list))

        return headers

    def encode_header(self, key, value_list):
        if not isinstance(value_list, list):
            value_list = [value_list]

        return [(key.encode(UTF8), self.encode_value(value)) for value in value_list]

    def encode_value(self, value):
        if isinstance(value, str):
            return value.encode(UTF8)
        return value

    async def serialize_content(self):
        if self.content is not None:
            if isinstance(self.content, bytes):
                body = self.content
            elif self.content_type == APPLICATION_JSON:
                body = json.dumps(self.content).encode(UTF8)
            else:
                body = str(self.content).encode(UTF8)
        else:
            body = b""
        return body


class JsonResponse(HttpResponse):
    def __init__(self, content=None, status_code=HTTPStatus.OK, headers=None):
        super().__init__(content, status_code, APPLICATION_JSON, headers)


class TemplateResponse(HttpResponse):
    def __init__(
        self,
        template_name=None,
        context=None,
        template_dir="templates",
        static_dir="static",
    ):
        super().__init__(None, HTTPStatus.OK, TEXT_HTML)
        self.template_name = template_name
        self.context = context or {}
        self.static_dir = static_dir
        self.template_dir = template_dir

    async def __call__(self, scope, receive, send):
        path_info = scope.get("path", "").lstrip("/") or scope.get(
            "raw_path", b""
        ).decode(UTF8).lstrip("/")

        if path_info.startswith("static/"):
            await self.handle_static_file(scope, receive, send)
        else:
            await self.render_template(scope, receive, send)

    async def render_template(self, scope, receive, send):
        if self.template_name is None:
            log.error("Template name is not provided.")
            not_found_response = JsonResponse(
                {"error": NOT_FOUND}, status_code=HTTPStatus.NOT_FOUND
            )
            await not_found_response(scope, receive, send)
            return

        template_path = os.path.join(self.template_dir, self.template_name)

        if not os.path.exists(template_path):
            log.error("Template not found:", template_path)
            not_found_response = JsonResponse(
                {"error": NOT_FOUND}, status_code=HTTPStatus.NOT_FOUND
            )
            await not_found_response(scope, receive, send)

        template_env = Environment(loader=FileSystemLoader(self.template_dir))
        template = template_env.get_template(self.template_name)
        content = template.render(**self.context)

        self.content = content.encode(UTF8)
        await super().__call__(scope, receive, send)

    async def handle_static_file(self, scope, receive, send):
        path_info = scope.get("path", "").lstrip("/") or scope.get(
            "raw_path", b""
        ).decode(UTF8).lstrip("/")
        static_prefix = "static/"

        if path_info.startswith(static_prefix):
            relative_path = path_info[len(static_prefix) :]
            file_path = os.path.join(self.static_dir, relative_path)

            if os.path.isfile(file_path):
                content_type, _ = mimetypes.guess_type(file_path)
                headers = [(b"content-type", content_type.encode(UTF8))]

                with open(file_path, "rb") as file:
                    body = file.read()

                await send(
                    {
                        "type": "http.response.start",
                        "status": HTTPStatus.OK,
                        "headers": headers,
                    }
                )

                await send(
                    {
                        "type": "http.response.body",
                        "body": body,
                        "more_body": False,
                    }
                )
            else:
                log.error("File not found:", file_path)

                # Return a 404 response for non-existing static files
                not_found_response = JsonResponse(
                    {"error": "Not Found"}, status_code=HTTPStatus.NOT_FOUND
                )
                await not_found_response(scope, receive, send)
        else:
            log.error("Unexpected path:", path_info)

            # Return a 404 response for unexpected paths
            not_found_response = JsonResponse(
                {"error": "Not Found"}, status_code=HTTPStatus.NOT_FOUND
            )
            await not_found_response(scope, receive, send)


class HttpResponseRedirect(HttpResponse):
    def __init__(self, url: str, status_code=HTTPStatus.FOUND, headers=None):
        super().__init__(content=None, status_code=status_code, headers=headers or {})
        self.headers["Location"] = url


class ForbiddenResponse(HttpResponse):
    def __init__(
        self,
        content=None,
        content_type=TEXT_PLAIN,
        status_code=HTTPStatus.FORBIDDEN,
        headers=None,
    ):
        super().__init__(content, status_code, content_type, headers)



import logging
from typing import Any, Callable, Dict

from inspira.helpers.error_templates import (
    format_method_not_allowed_exception,
    format_not_found_exception,
    format_internal_server_error,
    format_forbidden_exception,
    format_unauthorized_exception,
)


async def handle_method_not_allowed(
    scope: Dict[str, Any], receive: Callable, send: Callable
) -> None:
    method_not_allowed_response = format_method_not_allowed_exception()
    await method_not_allowed_response(scope, receive, send)


async def handle_not_found(
    scope: Dict[str, Any], receive: Callable, send: Callable
) -> None:
    not_found_response = format_not_found_exception()
    await not_found_response(scope, receive, send)


async def handle_forbidden(
    scope: Dict[str, Any], receive: Callable, send: Callable
) -> None:
    forbidden_response = format_forbidden_exception()
    await forbidden_response(scope, receive, send)


async def handle_unauthorized(
    scope: Dict[str, Any], receive: Callable, send: Callable
) -> None:
    unauthorized_response = format_unauthorized_exception()
    await unauthorized_response(scope, receive, send)


async def handle_internal_server_error(
    scope: Dict[str, Any], receive: Callable, send: Callable
) -> None:
    internal_server_error = format_internal_server_error()
    await internal_server_error(scope, receive, send)


async def default_error_handler(exc):
    logging.exception(exc)
    return format_internal_server_error()



from inspira.websockets import WebSocketControllerRegistry


def websocket(path: str):
    def decorator(cls):
        WebSocketControllerRegistry.register_controller(path, cls)
        return cls

    return decorator



UTF8 = "utf-8"

APPLICATION_JSON = "application/json"
TEXT_PLAIN = "text/plain"
TEXT_HTML = "text/html"

NOT_FOUND = "Not Found"

WEBSOCKET_SEND_TYPE = "websocket.send"
WEBSOCKET_ACCEPT_TYPE = "websocket.accept"
WEBSOCKET_CLOSE_TYPE = "websocket.close"
WEBSOCKET_RECEIVE_TYPE = "websocket.receive"
WEBSOCKET_DISCONNECT_TYPE = "websocket.disconnect"
WEBSOCKET_TYPE = "websocket"

SRC_DIRECTORY = "src"



from inspira.constants import TEXT_HTML
from inspira.responses import HttpResponse

template = """
<!DOCTYPE html>
<html>
<title>{title}</title>
<style>
html, body{{
  margin: 0;
  padding: 0;
  text-align: center;
  font-family: sans-serif;
}}

h1, a{{
  margin: 0;
  padding: 0;
  text-decoration: none;
}}

.section{{
  padding: 4rem 2rem;
}}

.section .error{{
  font-size: 150px;
}}

.page{{
  font-size: 20px;
  font-weight: 600;
}}

.back-home{{
  display: inline-block;
  text-transform: uppercase;
  font-weight: 600;
  padding: 0.75rem 1rem 0.6rem;
}}
</style>
<body>

<div class="section">
  <h1 class="error">{status_code}</h1>
  <div class="page">{message}</div>
</div>

</body>
</html>
"""


def format_internal_server_error() -> HttpResponse:
    msg = template.format(
        title="Internal Server Error",
        message="Internal Server Error"
        "<br><br>We are currently trying to fix the problem.",
        status_code=500,
    )
    return HttpResponse(content=msg, status_code=500, content_type=TEXT_HTML)


def format_not_found_exception() -> HttpResponse:
    msg = template.format(
        title="Not Found",
        message="Ooops!!! The page you are looking for is not found",
        status_code=404,
    )
    return HttpResponse(content=msg, status_code=404, content_type=TEXT_HTML)


def format_forbidden_exception() -> HttpResponse:
    msg = template.format(
        title="Forbidden",
        message="Forbidden",
        status_code=403,
    )
    return HttpResponse(content=msg, status_code=403, content_type=TEXT_HTML)


def format_unauthorized_exception() -> HttpResponse:
    msg = template.format(
        title="Unauthorized",
        message="Unauthorized",
        status_code=401,
    )
    return HttpResponse(content=msg, status_code=401, content_type=TEXT_HTML)


def format_method_not_allowed_exception() -> HttpResponse:
    msg = template.format(
        title="Method Not Allowed",
        message="Method Not Allowed"
        "<br><br>The method is not allowed for the requested URL.",
        status_code=405,
    )
    return HttpResponse(content=msg, status_code=405, content_type=TEXT_HTML)



import os

import click

from inspira.cli.create_app import generate_project
from inspira.cli.init_file import create_init_file
from inspira.constants import SRC_DIRECTORY
from inspira.utils import singularize


def create_src_directory():
    app_file_path = "main.py"

    if not os.path.exists(app_file_path):
        generate_project()
    if not os.path.exists(SRC_DIRECTORY):
        os.makedirs(SRC_DIRECTORY)
        create_init_file(SRC_DIRECTORY)


def create_test_directory(controller_directory):
    test_directory = os.path.join(controller_directory, "tests")
    os.makedirs(test_directory)

    create_init_file(test_directory)


def create_controller_file(name, is_websocket):
    controller_directory = os.path.join(SRC_DIRECTORY, name)
    singularize_name = singularize(name.lower())

    os.makedirs(controller_directory)

    create_test_directory(controller_directory)

    controller_template_file = "controller_template.txt"
    create_init_file(controller_directory)

    controller_file = os.path.join(
        controller_directory, f"{singularize_name}_controller.py"
    )

    if is_websocket:
        controller_template_file = "websocket_controller_template.txt"

    template_path = os.path.join(
        os.path.dirname(__file__), "templates", controller_template_file
    )

    with open(template_path, "r") as template_file, open(
        controller_file, "w"
    ) as output_file:
        content = (
            template_file.read()
            .replace("{{controller_name}}", singularize_name.capitalize())
            .replace("{{root_path}}", name.lower())
        )
        output_file.write(content)

    click.echo(f"Module '{singularize_name}' created successfully.")



import bcrypt

from inspira.constants import UTF8


class UserMixin:
    def __init__(self):
        self.password = None

    @property
    def is_active(self):
        return True

    @property
    def is_authenticated(self):
        return self.is_active

    @property
    def is_anonymous(self):
        return False

    def set_password(self, password):
        hashed_password = bcrypt.hashpw(password.encode(UTF8), bcrypt.gensalt())
        self.password = hashed_password.decode(UTF8)

    def check_password_hash(self, password):
        return bcrypt.checkpw(password.encode(UTF8), self.password.encode(UTF8))


class AnonymousUserMixin:
    @property
    def is_authenticated(self):
        return False

    @property
    def is_active(self):
        return False

    @property
    def is_anonymous(self):
        return True



import inspect


def convert_param_type(value, param_type):
    try:
        if param_type is None or param_type == inspect.Parameter.empty:
            return str(value)
        return param_type(value)
    except ValueError:
        return value



import inspect
from typing import Any, Dict

from inspira.requests import Request
from inspira.utils.param_converter import convert_param_type


async def invoke_handler(handler, request: Request, scope: Dict[str, Any], params=None):
    handler_signature = inspect.signature(handler)
    handler_params = {}
    for param_name, param in handler_signature.parameters.items():
        if param_name == "request":
            handler_params["request"] = request
        elif param_name == "scope":
            handler_params["scope"] = scope
        elif param_name in params:
            handler_params[param_name] = convert_param_type(
                params[param_name], param.annotation
            )
        elif param.default != inspect.Parameter.empty:
            handler_params[param_name] = param.default
        else:
            handler_params[param_name] = None

    return await handler(**handler_params)



import ast


def parse_controller_decorators(file_path: str) -> bool:
    with open(file_path, "r") as file:
        tree = ast.parse(file.read(), filename=file_path)

    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            for decorator in node.decorator_list:
                if (
                    isinstance(decorator, ast.Call)
                    and isinstance(decorator.func, ast.Name)
                    and (
                        decorator.func.id == "path" or decorator.func.id == "websocket"
                    )
                ):
                    return True
    return False



from httpx import AsyncClient

from inspira.enums import HttpMethod


class TestClient:
    __test__ = False

    def __init__(self, app):
        self.app = app

    async def request(self, method, path, **kwargs):
        async with AsyncClient(app=self.app, base_url="http://testserver") as client:
            return await getattr(client, method.lower())(path, **kwargs)

    async def get(self, path, **kwargs):
        return await self.request(HttpMethod.GET.value, path, **kwargs)

    async def post(self, path, **kwargs):
        return await self.request(HttpMethod.POST.value, path, **kwargs)

    async def put(self, path, **kwargs):
        return await self.request(HttpMethod.PUT.value, path, **kwargs)

    async def delete(self, path, **kwargs):
        return await self.request(HttpMethod.DELETE.value, path, **kwargs)

    async def patch(self, path, **kwargs):
        return await self.request(HttpMethod.PATCH.value, path, **kwargs)

    async def options(self, path, **kwargs):
        return await self.request(HttpMethod.OPTIONS.value, path, **kwargs)



from typing import Dict, Any, Callable

from inspira.auth.auth_utils import decode_auth_token
from inspira.auth.mixins.user_mixin import AnonymousUserMixin
from inspira.globals import get_global_app
from inspira.helpers.error_handlers import handle_forbidden
from inspira.logging import log
from inspira.requests import RequestContext
from inspira.utils.session_utils import (
    decode_session_data,
    get_session_token_from_request,
)


class UserLoaderMiddleware:
    def __init__(self, user_model):
        self.user_model = user_model
        self.app = get_global_app()

    async def __call__(self, handler):
        async def middleware(scope: Dict[str, Any], receive: Callable, send: Callable):
            request = RequestContext().get_request()
            session_cookie = get_session_token_from_request(
                request, self.app.config["SESSION_COOKIE_NAME"]
            )

            user = None

            if session_cookie:
                decoded_session = decode_session_data(
                    session_cookie, self.app.secret_key
                )
                if decoded_session is None:
                    log.error("Invalid session format.")
                    return await handle_forbidden(scope, receive, send)
                else:
                    user_id = decode_auth_token(decoded_session.get("token", None))

                    if user_id:
                        user = self.user_model.query.get(user_id)
            RequestContext.set_current_user(user or AnonymousUserMixin())
            request.user = RequestContext.get_current_user()

            await handler(scope, receive, send)

        return middleware



import json
import urllib.parse
from typing import Any, Callable, Dict

from inspira.constants import UTF8


class RequestContext:
    _current_request = None
    _current_user = None

    @classmethod
    def set_request(cls, request):
        cls._current_request = request

    @classmethod
    def get_request(cls):
        return cls._current_request

    @classmethod
    def get_current_user(cls):
        return cls._current_user

    @classmethod
    def set_current_user(cls, user):
        cls._current_user = user


class Request:
    def __init__(self, scope: Dict[str, Any], receive: Callable, send: Callable):
        self.scope = scope
        self.receive = receive
        self.send = send
        self._session = {}
        self._headers = {}
        self._forbidden = False
        self.user = None

    def is_forbidden(self):
        return self._forbidden

    def set_forbidden(self):
        self._forbidden = True

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, value):
        self._session = value

    def set_session(self, key, value):
        if not isinstance(self._session, dict):
            self._session = {}

        self._session[key] = value

    def get_session(self, key, default=None):
        if not isinstance(self._session, dict):
            return default

        return self._session.get(key, default)

    def remove_session(self, key, default=None):
        if not isinstance(self._session, dict):
            return default

        return self._session.pop(key, default)

    def get_headers(self):
        return dict(
            (key.decode(UTF8), value.decode(UTF8))
            for key, value in self.scope.get("headers", [])
        )

    def set_header(self, key, value):
        self._headers[key] = value

    def get_request_headers(self):
        return list(
            (key.encode(UTF8), value.encode(UTF8))
            for key, value in self._headers.items()
        )

    def cookies(self):
        cookie_header = self.get_headers().get("cookie", "")
        if cookie_header:
            try:
                return {
                    key.strip(): value.strip()
                    for key, value in [
                        cookie.split("=", 1) for cookie in cookie_header.split(";")
                    ]
                }
            except ValueError:
                return {}
        return {}

    async def json(self):
        body = await self._get_body()
        if body:
            return json.loads(body.decode(UTF8))
        return {}

    async def _get_boundary(self):
        content_type_header = self.get_headers().get("content-type", "")
        if "multipart/form-data" in content_type_header:
            parts = content_type_header.split(";")
            for part in parts:
                if "boundary" in part:
                    _, boundary = part.strip().split("=")
                    return boundary.strip('"')
        return None

    async def form(self):
        content_type_header = self.get_headers().get("content-type", "")
        if "application/x-www-form-urlencoded" in content_type_header:
            body = await self._get_body()
            form_data = urllib.parse.parse_qsl(body.decode(UTF8))
            return {key: value for key, value in form_data}
        elif "multipart/form-data" in content_type_header:
            # Handle multipart form data
            return await self._parse_multipart_form_data()

        return {}

    async def _get_body(self):
        body = b""
        more_body = True
        while more_body:
            message = await self.receive()
            body += message.get("body", b"")
            more_body = message.get("more_body", False)
        return body

    async def _parse_multipart_form_data(self):
        form_data = await self._get_body()
        boundary = await self._get_boundary()

        if not boundary or not form_data:
            return {}

        parts = form_data.split(b"--" + boundary.encode())

        form_data_dict = {}
        for part in parts:
            if not part.strip():
                continue

            headers_content = part.split(b"\r\n\r\n", 1)
            if len(headers_content) == 2:
                headers, content = headers_content
                content_lines = content.split(b"\r\n")
                # Filter out empty lines and join the rest
                content = b"\r\n".join(line for line in content_lines if line.strip())
            else:
                headers, content = headers_content[0], b""

            headers_lines = headers.decode(UTF8).split("\r\n")

            content_disposition = next(
                (
                    line
                    for line in headers_lines
                    if line.startswith("Content-Disposition")
                ),
                None,
            )

            if not content_disposition:
                continue

            _, name = content_disposition.split(";")
            name = name.split("=")[1].strip('"')

            content_str = content.decode(UTF8)

            form_data_dict[name] = content_str.strip()

        return form_data_dict



import inspect
from typing import Any, Callable, List, Optional, Type


def resolve_dependencies_automatic(cls: Type) -> Optional[List[Any]]:
    dependencies: List[Any] = []

    # Get the constructor parameters, if available
    init_method = getattr(cls, "__init__", None)

    if init_method and init_method != object.__init__:
        constructor_params = inspect.signature(init_method).parameters.values()
        for param in constructor_params:
            if param.name != "self":  # Exclude 'self' parameter
                param_type = param.annotation
                dependency = resolve_dependency(param_type)
                dependencies.append(dependency)

    return dependencies if dependencies else None


def resolve_dependency(dependency_type: Type[Callable]) -> Optional[Callable]:
    if dependency_type:
        init_method = getattr(dependency_type, "__init__", None)
        if init_method and init_method != object.__init__:
            constructor_params = inspect.signature(init_method).parameters.values()
            resolved_params = [
                resolve_dependency(param.annotation)
                for param in constructor_params
                if param.name != "self"
            ]
            return dependency_type(*resolved_params)
        else:
            return dependency_type()
    else:
        return None



from inspira.constants import WEBSOCKET_DISCONNECT_TYPE, WEBSOCKET_RECEIVE_TYPE
from inspira.logging import log
from inspira.utils.dependency_resolver import resolve_dependencies_automatic
from inspira.websockets import WebSocket, WebSocketControllerRegistry


async def handle_websocket(scope, receive, send):
    path = scope["path"]
    controller_cls = WebSocketControllerRegistry.get_controller(path)

    if not controller_cls:
        log.info(f"No WebSocket controller registered for path: {path}")
        return

    websocket_cls = WebSocket(scope, receive, send)

    dependencies = resolve_dependencies_automatic(controller_cls)
    instance = (
        controller_cls(*dependencies) if dependencies is not None else controller_cls()
    )

    try:
        await instance.on_open(websocket_cls)

        while True:
            message = await websocket_cls.receive()

            if message["type"] == WEBSOCKET_DISCONNECT_TYPE:
                break

            if message["type"] == WEBSOCKET_RECEIVE_TYPE:
                await instance.on_message(websocket_cls, message)

    except Exception as e:
        log.info(f"WebSocket connection error: {e}")
    finally:
        await instance.on_close(websocket_cls)



from typing import Literal

import inflect

p = inflect.engine()


def singularize(word: str) -> str | Literal[False]:
    return p.singular_noun(word) or word


def pluralize_word(word: str) -> str:
    if word.endswith("s"):
        return word
    return p.plural(word) or word



import logging

log = logging.getLogger("Inspira")

handler = logging.StreamHandler()
formatter = logging.Formatter("[%(asctime)s] %(levelname)s in %(module)s: %(message)s")
handler.setFormatter(formatter)
log.addHandler(handler)
log.setLevel(logging.INFO)



def default_max_age():
    return 24 * 60 * 60 * 31


class Config:
    def __init__(self):
        self.config_data = {
            "SESSION_COOKIE_NAME": "session",
            "SESSION_MAX_AGE": default_max_age(),
            "SESSION_COOKIE_DOMAIN": None,
            "SESSION_COOKIE_PATH": None,
            "SESSION_COOKIE_HTTPONLY": True,
            "SESSION_COOKIE_SECURE": True,
            "SESSION_COOKIE_SAMESITE": None,
            "TOKEN_EXPIRATION_TIME": 3600,
            "SECRET_KEY": "change_me",
        }

    def __getitem__(self, key):
        return self.config_data.get(key, None)

    def __setitem__(self, key, value):
        self.config_data[key] = value



from typing import Callable, Type


def path(path: str) -> Callable[[Type], Type]:
    def decorator(cls: Type) -> Type:
        cls.__path__ = path
        cls.__is_controller__ = True
        return cls

    return decorator



from datetime import datetime, timedelta

import jwt

from inspira.requests import RequestContext
from inspira.globals import get_global_app


app = get_global_app()

if app:
    SECRET_KEY = app.secret_key
    TOKEN_EXPIRATION_TIME = app.config["TOKEN_EXPIRATION_TIME"]
else:
    SECRET_KEY = "dummy"
    TOKEN_EXPIRATION_TIME = 3600


def login_user(user_id):
    token = encode_auth_token(user_id)
    request = RequestContext.get_request()
    request.set_session("token", token)


def logout_user():
    request = RequestContext.get_request()
    session = request.session

    if session and "token" in session:
        request.remove_session("token")


def encode_auth_token(user_id):
    payload = {
        "iat": datetime.utcnow(),
        "exp": datetime.utcnow() + timedelta(seconds=TOKEN_EXPIRATION_TIME),
        "sub": user_id,
    }
    return jwt.encode(payload, SECRET_KEY, algorithm="HS256")


def decode_auth_token(token):
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=["HS256"])
        return payload["sub"]
    except jwt.ExpiredSignatureError:
        return None
    except jwt.InvalidTokenError:
        return None



from .dependency_resolver import resolve_dependencies_automatic, resolve_dependency
from .inflection import pluralize_word, singularize
from .naming import convert_to_camel_case, convert_to_snake_case
from .secret_key import get_random_secret_key



import sys

import click
from sqlalchemy import select, create_engine, inspect, Index
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import declarative_base, scoped_session, sessionmaker
from sqlalchemy.sql.ddl import CreateTable, CreateIndex, DropIndex
from sqlalchemy.sql.expression import func
import os
from sqlalchemy import MetaData, Column, Integer, String, text

from inspira.logging import log
from inspira.migrations.utils import (
    get_or_create_migration_directory,
    get_migration_files,
    load_model_file,
    get_columns_from_model,
    generate_add_column_sql,
    generate_drop_column_sql,
    generate_migration_file_for_create_table,
    generate_rename_column_sql,
    generate_empty_sql_file,
    get_indexes_from_model,
    generate_migration_file,
    migration_file_exist,
)

PROJECT_ROOT = os.path.abspath(".")
sys.path.append(PROJECT_ROOT)

try:
    from database import Base, engine, db_session
except ImportError:
    Base = declarative_base()
    engine = create_engine("sqlite:///:memory:")
    db_session = scoped_session(
        sessionmaker(autocommit=False, autoflush=False, bind=engine)
    )


class Migration(Base):
    __tablename__ = "migrations"
    id = Column(Integer, primary_key=True)
    migration_name = Column(String(255))
    version = Column(Integer)


def initialize_database(engine):
    Base.metadata.create_all(engine)


def execute_sql_file(file_path):
    with open(file_path, "r") as file:
        sql_content = file.read()

    sql_statements = [
        statement.strip() for statement in sql_content.split(";") if statement.strip()
    ]

    with engine.connect() as connection:
        try:
            for statement in sql_statements:
                connection.execute(text(statement))
            connection.commit()
            log.info("Migration run successfully.")
        except SQLAlchemyError as e:
            log.error("Error:", e)
            connection.rollback()
            log.info("Transaction rolled back.")


def generate_create_table_sql(model_name):
    metadata = Base.metadata
    table = metadata.tables[model_name]
    sql = str(CreateTable(table).compile(engine))[:-2] + ";" + "\n"
    index_sqls = [CreateIndex(index).compile(engine) for index in table.indexes]

    for index_sql in index_sqls:
        sql += "\n" + str(index_sql) + ";"

    return sql


def create_migrations(entity_name, empty_migration_file):
    if empty_migration_file:
        generate_empty_sql_file(entity_name, empty_migration_file)

    module = load_model_file(entity_name)

    if module is None:
        log.error(f"Module '{entity_name}' not found.")
        return

    model = getattr(module, module.__name__)

    if not get_existing_columns(entity_name):
        generate_migration_file_for_create_table(
            generate_create_table_sql(entity_name), entity_name
        )
        return

    handle_columns(entity_name, model)
    handle_indexes(entity_name, model)


def handle_columns(entity_name, model):
    existing_columns = get_existing_columns(entity_name)
    new_columns = get_columns_from_model(model)

    renamed_columns = [
        (old_col, new_col.key)
        for old_col, new_col in zip(existing_columns, new_columns)
        if old_col != new_col.key
    ]

    if renamed_columns:
        generate_rename_column_sql(entity_name, existing_columns, new_columns)
        return

    added_columns = [col for col in new_columns if col.key not in existing_columns]
    if added_columns:
        generate_add_column_sql(entity_name, existing_columns, added_columns)
        return

    removed_columns = [col for col in existing_columns if col not in new_columns]
    if removed_columns:
        generate_drop_column_sql(entity_name, existing_columns, new_columns)


def handle_indexes(entity_name, model):
    existing_indexes = get_existing_indexes(entity_name)
    new_indexes = get_indexes_from_model(model)

    generate_add_index_sql(entity_name, existing_indexes, new_indexes)
    generate_drop_index_sql(entity_name, existing_indexes, new_indexes)


def run_migrations(module_name):
    with engine.connect() as connection:
        if not engine.dialect.has_table(connection, "migrations"):
            initialize_database(engine)

        if not engine.dialect.has_table(connection, module_name):
            initialize_database(engine)

        migration_dir = get_or_create_migration_directory(module_name)
        migration_files = get_migration_files(migration_dir)

        for file in migration_files:
            migration_name = os.path.basename(file).replace(".sql", "")

            current_version = (
                connection.execute(
                    select(func.max(Migration.version)).where(
                        Migration.migration_name == migration_name
                    )
                ).scalar()
                or 0
            )

            if not current_version:
                execute_sql_file(file)
                click.echo(
                    f"Applying migration for {migration_name} version {current_version}"
                )

                insert_migration(current_version, migration_name)


def get_existing_columns(table_name):
    metadata = MetaData()
    metadata.reflect(bind=engine)

    if table_name in metadata.tables:
        table = metadata.tables[table_name]
        return [column.name for column in table.columns]
    else:
        return None


def insert_migration(current_version, migration_name):
    migration = Migration()
    migration.version = current_version + 1
    migration.migration_name = migration_name
    db_session.add(migration)
    db_session.commit()


def get_existing_indexes(table_name):
    inspector = inspect(engine)
    indexes = inspector.get_indexes(table_name)
    return [index["name"] for index in indexes]


def generate_add_index_sql(table_name, existing_indexes, new_indexes):
    for new_index in new_indexes:
        if new_index.name not in existing_indexes:
            index_sql = str(CreateIndex(new_index).compile(engine)) + ";"

            migration_file_name = f"add_index_{new_index.name}"

            if migration_file_exist(table_name, migration_file_name):
                return

            generate_migration_file(table_name, index_sql, migration_file_name)


def generate_drop_index_sql(table_name, existing_indexes, new_indexes):
    new_index_names = set(index.name for index in new_indexes)

    for removed_index_name in set(existing_indexes) - new_index_names:
        removed_index = Index(removed_index_name, text("dummy"))
        drop_index_sql = str(DropIndex(removed_index).compile(engine)) + ";"

        migration_file_name = f"drop_index_{removed_index_name}"

        if migration_file_exist(table_name, migration_file_name):
            return

        generate_migration_file(table_name, drop_index_sql, migration_file_name)



from typing import Callable, Type

from inspira.enums import HttpMethod


def get(path: str = "") -> Callable[[Type], Type]:
    def decorator(handler: Type) -> Type:
        handler.__method__ = HttpMethod.GET
        handler.__path__ = path
        handler.__is_handler__ = True
        return handler

    return decorator


def post(path: str = "") -> Callable[[Type], Type]:
    def decorator(handler: Type) -> Type:
        handler.__method__ = HttpMethod.POST
        handler.__path__ = path
        handler.__is_handler__ = True
        return handler

    return decorator


def put(path: str = "") -> Callable[[Type], Type]:
    def decorator(handler: Type) -> Type:
        handler.__method__ = HttpMethod.PUT
        handler.__path__ = path
        handler.__is_handler__ = True
        return handler

    return decorator


def patch(path: str = "") -> Callable[[Type], Type]:
    def decorator(handler: Type) -> Type:
        handler.__method__ = HttpMethod.PATCH
        handler.__path__ = path
        handler.__is_handler__ = True
        return handler

    return decorator


def delete(path: str = "") -> Callable[[Type], Type]:
    def decorator(handler: Type) -> Type:
        handler.__method__ = HttpMethod.DELETE
        handler.__path__ = path
        handler.__is_handler__ = True
        return handler

    return decorator



from typing import Any, Callable, Dict

from inspira.requests import Request
from inspira.responses import TemplateResponse


async def handle_static_files(
    scope: Dict[str, Any], receive: Callable, send: Callable, request: Request
) -> None:
    template_response = TemplateResponse(request, scope["path"])
    await template_response(scope, receive, send)



import os

import click

from inspira.utils import get_random_secret_key


def generate_project():
    create_app_file()
    click.echo("Project created successfully")


def create_app_file():
    template_path = os.path.join(
        os.path.dirname(__file__), "templates", "app_template.txt"
    )
    output_path = "main.py"

    with open(template_path, "r") as template_file, open(
        output_path, "w"
    ) as output_file:
        content = template_file.read().replace(
            "{{secret_key}}", get_random_secret_key()
        )
        output_file.write(content)



from typing import List, Dict, Any, Callable

from inspira.helpers.error_handlers import handle_forbidden
from inspira.requests import Request


class CORSMiddleware:
    def __init__(
        self,
        allow_origins: List[str] = None,
        allow_credentials: bool = False,
        allow_methods: List[str] = None,
        allow_headers: List[str] = None,
    ):
        self.allow_origins = allow_origins or []
        self.allow_credentials = allow_credentials
        self.allow_methods = allow_methods or []
        self.allow_headers = allow_headers or []

    async def __call__(self, handler):
        async def middleware(scope: Dict[str, Any], receive: Callable, send: Callable):
            request = Request(scope, receive, send)
            origin = request.get_headers().get("origin")

            if scope["method"] == "OPTIONS":
                return await self.handle_options(scope, receive, send, origin)

            if origin is not None and not self.is_origin_allowed(origin):
                return await handle_forbidden(scope, receive, send)

            async def send_wrapper(message):
                if message["type"] == "http.response.start":
                    headers = message.get("headers", [])

                    if origin in self.allow_origins:
                        headers.extend(self.get_cors_headers(origin))

                    message["headers"] = headers

                await send(message)

            await handler(scope, receive, send_wrapper)

        return middleware

    def is_origin_allowed(self, origin):
        return origin is not None and (
            origin in self.allow_origins or "*" in self.allow_origins
        )

    def get_cors_headers(self, origin):
        return [
            (b"Access-Control-Allow-Origin", origin.encode()),
            (
                b"Access-Control-Allow-Credentials",
                str(self.allow_credentials).lower().encode(),
            ),
            (b"Access-Control-Allow-Methods", ",".join(self.allow_methods).encode()),
            (b"Access-Control-Allow-Headers", ",".join(self.allow_headers).encode()),
        ]

    async def handle_options(self, scope, receive, send, origin):
        headers = self.get_cors_headers(origin)
        headers.append(
            (
                b"Access-Control-Allow-Headers",
                ",".join(self.allow_headers + ["Content-Type"]).encode(),
            )
        )
        response_message = {
            "type": "http.response.start",
            "status": 200,
            "headers": headers,
        }

        await send(response_message)
        await send({"type": "http.response.body", "body": b""})



from typing import Type


class WebSocketControllerRegistry:
    _controllers = {}

    @classmethod
    def register_controller(cls, path: str, controller_cls: Type):
        cls._controllers[path] = controller_cls
        return controller_cls

    @classmethod
    def get_controller(cls, path: str):
        return cls._controllers.get(path)



import importlib
import os

from sqlalchemy import String, Integer

from inspira.cli.create_controller import create_init_file
from inspira.constants import SRC_DIRECTORY
from inspira.logging import log
from inspira.utils import singularize


def get_or_create_migration_directory(name: str):
    controller_directory = os.path.join(SRC_DIRECTORY, name)
    migration_directory = os.path.join(controller_directory, "migrations")

    if not os.path.exists(migration_directory):
        log.error(f"Module '{name}' doesn't exists.")
        return

    os.makedirs(migration_directory, exist_ok=True)

    create_init_file(migration_directory)

    return migration_directory


def load_model_file(entity_name):
    module_path = os.path.join(
        SRC_DIRECTORY, entity_name.replace(".", "/" + entity_name)
    )
    model_file_path = os.path.join(module_path, f"{singularize(entity_name)}.py")
    model_name = singularize(entity_name).capitalize()
    spec = importlib.util.spec_from_file_location(model_name, model_file_path)
    module = importlib.util.module_from_spec(spec)

    try:
        spec.loader.exec_module(module)
    except FileNotFoundError:
        return None

    return module


def generate_drop_column_sql(table_name, existing_columns, new_columns):
    sql_statements = ""
    migration_name = ""

    for i, col in enumerate(existing_columns):
        if col not in new_columns:
            add_underscore = "_" if i < len(new_columns) - 1 else ""
            sql_statements += f"ALTER TABLE {table_name} DROP COLUMN {col};\n"
            migration_name += col + add_underscore

    if sql_statements:
        migration_file_name = f"drop_column_{migration_name}"
        if migration_file_exist(table_name, migration_file_name):
            return

        generate_migration_file(table_name, sql_statements, migration_file_name)


def generate_add_column_sql(table_name, existing_columns, new_columns):
    sql_statements = ""
    migration_name = ""
    for i, col in enumerate(new_columns):
        if col.key not in existing_columns:
            add_underscore = "_" if i < len(new_columns) - 1 else ""
            sql_statements += f"ALTER TABLE {table_name} ADD COLUMN {generate_column_sql(col).strip()};\n"
            migration_name += col.key + add_underscore

    if sql_statements:
        migration_file_name = f"add_column_{migration_name}"
        if migration_file_exist(table_name, migration_file_name):
            return
        generate_migration_file(table_name, sql_statements, migration_file_name)


def generate_rename_column_sql(table_name, existing_columns, new_columns):
    sql_statements = ""
    migration_name = ""

    for i, (old_col, new_col) in enumerate(zip(existing_columns, new_columns)):
        if old_col != new_col.key:
            add_underscore = "_" if i > 0 else ""
            sql_statements += (
                f"ALTER TABLE {table_name} RENAME COLUMN {old_col} TO {new_col.key};"
            )
            migration_name += f"{old_col}_to_{new_col.key}{add_underscore}"

    if sql_statements:
        migration_file_name = f"rename_column_{migration_name}"
        if migration_file_exist(table_name, migration_file_name):
            return

        generate_migration_file(table_name, sql_statements, migration_file_name)


def generate_migration_file_for_create_table(sql_str, table_name):
    migration_file_name = f"create_table_{table_name}"

    if migration_file_exist(table_name, migration_file_name):
        return

    generate_migration_file(table_name, sql_str, migration_file_name)


def migration_file_exist(table_name: str, migration_file_name: str) -> bool:
    migration_dir = get_or_create_migration_directory(table_name)
    migration_files = get_migration_files(migration_dir)

    migration_exists = any(
        migration_file_name in migration for migration in migration_files
    )
    if migration_exists:
        log.info(f"Migration {migration_file_name} already exists. Skipping...")

    return migration_exists


def generate_empty_sql_file(module, migration_name):
    generate_migration_file(module, "", migration_name)


def get_migration_files(migration_dir):
    migration_files = [
        os.path.join(migration_dir, f)
        for f in os.listdir(migration_dir)
        if f.endswith(".sql")
        and f.split("_")[0].isdigit()
        and os.path.isfile(os.path.join(migration_dir, f))
    ]
    migration_files.sort()
    return migration_files


def get_latest_migration_number(migration_dir):
    migration_files = [f for f in os.listdir(migration_dir) if f.endswith(".sql")]
    if not migration_files:
        return 0

    latest_migration = max(int(f.split("_")[0]) for f in migration_files)
    return latest_migration


def generate_migration_file(module_name, migration_sql, migration_name):
    migration_dir = get_or_create_migration_directory(module_name)
    latest_migration_number = get_latest_migration_number(migration_dir)
    new_migration_number = latest_migration_number + 1
    migration_file_path = os.path.join(
        migration_dir, f"{str(new_migration_number).zfill(4)}_{migration_name}.sql"
    )

    with open(migration_file_path, "w") as migration_file:
        migration_file.write(migration_sql)

    log.info(
        f"Migration file '{str(new_migration_number).zfill(4)}_{migration_name}.sql' created."
    )
    return migration_file


def generate_column_sql(column):
    column_name = column.key
    constraints = []

    if isinstance(column.type, String):
        column_type = f"VARCHAR({column.type.length})"
    elif isinstance(column.type, Integer):
        column_type = "INTEGER"
    else:
        column_type = str(column.type)

    if column.primary_key:
        column_type += " PRIMARY KEY"

    if column.nullable:
        column_type += " NULL"
    else:
        column_type += " NOT NULL"

    if column.unique:
        constraints.append("UNIQUE")

    return f"{column_name} {column_type} {''.join(constraints)}"


def get_columns_from_model(model_class):
    return model_class.__table__.columns


def get_indexes_from_model(model_class):
    return model_class.__table__.indexes


def get_all_module_names():
    module_names = []

    for dir_entry in os.listdir(SRC_DIRECTORY):
        full_path = os.path.join(SRC_DIRECTORY, dir_entry)
        if os.path.isdir(full_path):
            migrations_dir = os.path.join(full_path, "migrations")
            if os.path.exists(migrations_dir):
                module_names.append(dir_entry)

    return module_names



import datetime
import uuid
from typing import Dict, Any, Callable

from inspira.globals import get_global_app
from inspira.helpers.error_handlers import handle_forbidden
from inspira.inspira import RequestContext
from inspira.logging import log
from inspira.utils.session_utils import (
    encode_session_data,
    decode_session_data,
    get_session_token_from_request,
)


class SessionMiddleware:
    def __init__(self):
        self.app = get_global_app()

    def build_set_cookie_header(self, session_data):
        encoded_payload = encode_session_data(session_data, self.app.secret_key)
        expires_date = datetime.datetime.utcnow() + datetime.timedelta(
            seconds=self.app.config["SESSION_MAX_AGE"]
        )
        formatted_expires = expires_date.strftime("%a, %d %b %Y %H:%M:%S GMT")

        cookie_value = (
            f"{self.app.config['SESSION_COOKIE_NAME']}={encoded_payload}; "
            f"Expires={formatted_expires}; Path={self.app.config['SESSION_COOKIE_PATH'] or '/'}; HttpOnly"
        )

        if self.app.config["SESSION_COOKIE_DOMAIN"]:
            cookie_value += f"; Domain={self.app.config['SESSION_COOKIE_DOMAIN']}"

        if self.app.config["SESSION_COOKIE_SECURE"]:
            cookie_value += "; Secure"

        if self.app.config["SESSION_COOKIE_SAMESITE"]:
            cookie_value += f"; SameSite={self.app.config['SESSION_COOKIE_SAMESITE']}"

        return cookie_value

    async def __call__(self, handler):
        async def middleware(scope: Dict[str, Any], receive: Callable, send: Callable):
            async def send_wrapper(message):
                if message["type"] == "http.response.start":
                    headers = message.get("headers", [])
                    request = RequestContext().get_request()
                    session_cookie = get_session_token_from_request(
                        request, self.app.config["SESSION_COOKIE_NAME"]
                    )

                    decoded_session = {}

                    if request.session and not session_cookie:
                        request.set_session("session_id", str(uuid.uuid4()))

                    if session_cookie:
                        decoded_session = decode_session_data(
                            session_cookie, self.app.secret_key
                        )

                        if decoded_session is None:
                            log.error("Invalid session format.")
                            return await handle_forbidden(scope, receive, send)

                    if not request.session or decoded_session != request.session:
                        if request.session:
                            cookie_value = self.build_set_cookie_header(request.session)

                            headers.append((b"Set-Cookie", cookie_value.encode()))
                        else:
                            cookie_value = f"{self.app.config['SESSION_COOKIE_NAME']}=; Expires=Thu, 01 Jan 1970 00:00:00 GMT; Path={self.app.config['SESSION_COOKIE_PATH'] or '/'}; HttpOnly"

                            headers.append((b"Set-Cookie", cookie_value.encode()))

                    message["headers"] = headers

                await send(message)

            await handler(scope, receive, send_wrapper)

        return middleware



import os

import click


def create_database_file(database_name, database_type):
    template_path = os.path.join(
        os.path.dirname(__file__), "templates", "database_template.txt"
    )
    output_path = "database.py"

    if database_type == "postgres":
        database_url = f"postgresql://USERNAME:PASSWORD@localhost:5432/{database_name}"
    elif database_type == "mysql":
        database_url = f"mysql://USERNAME:PASSWORD@localhost:3306/{database_name}"
    elif database_type == "sqlite":
        database_url = f"sqlite:///{database_name}.db"
    elif database_type == "mssql":
        # Assuming a Windows authentication connection
        database_url = (
            f"mssql+pyodbc://"
            f"@localhost/{database_name}?driver=ODBC+Driver+17+for+SQL+Server"
        )

    with open(template_path, "r") as template_file, open(
        output_path, "w"
    ) as output_file:
        template_content = template_file.read()
        template_content = template_content.replace("{{database_url}}", database_url)
        output_file.write(template_content)

    click.echo("Database file created successfully.")



import os
import re

from inspira.constants import SRC_DIRECTORY
from inspira.utils import pluralize_word, singularize


def generate_service_file(module_name):
    model_directory = os.path.join(SRC_DIRECTORY, module_name.lower())

    template_path = os.path.join(
        os.path.dirname(__file__), "templates", "service_template.txt"
    )
    repository_file_path = os.path.join(
        model_directory, f"{singularize(module_name.lower())}_service.py"
    )

    with open(template_path, "r") as template_file, open(
        repository_file_path, "w"
    ) as output_file:
        content = (
            template_file.read()
            .replace("{{model_name}}", pluralize_word(module_name))
            .replace("{{model_name_upper}}", singularize(module_name.capitalize()))
            .replace("{{module_name}}", module_name)
            .replace("{{model_name_singular}}", singularize(module_name.lower()))
        )
        output_file.write(content)

    add_service_dependency_to_controller(module_name)


def add_service_dependency_to_controller(module_name):
    controller_file_path = os.path.join(
        SRC_DIRECTORY,
        module_name.lower(),
        f"{singularize(module_name.lower())}_controller.py",
    )
    module_name_capitalized = singularize(module_name.lower()).capitalize()
    singularized_module_name = singularize(module_name.lower())

    with open(controller_file_path, "r") as controller_file:
        existing_content = controller_file.read()

    # Define the __init__ method
    init_method = f"""\n
    def __init__(self, {singularized_module_name}_service: {module_name_capitalized}Service):
        self._{singularized_module_name}_service = {singularized_module_name}_service"""

    import_statement = (
        f""
        f"\n\nfrom src.{module_name.lower()}.{singularize(module_name.lower())}_service"
        f" import {module_name_capitalized}Service\n\n\n"
    )

    # Update the regular expression to match @path or @websocket
    class_pattern = re.compile(
        r"@(path|websocket)\(.+\)\nclass\s+[A-Za-z_][A-Za-z0-9_]*:"
    )
    match = class_pattern.search(existing_content)

    if match:
        start, end = match.span()
        updated_content = (
            existing_content[:start].rstrip()
            + import_statement
            + existing_content[start:end]
            + init_method
            + existing_content[end:]
        )

        with open(controller_file_path, "w") as controller_file:
            controller_file.write(updated_content)



import json

from inspira.constants import (
    WEBSOCKET_ACCEPT_TYPE,
    WEBSOCKET_CLOSE_TYPE,
    WEBSOCKET_SEND_TYPE,
    WEBSOCKET_TYPE,
)
from inspira.logging import log


class WebSocket:
    def __init__(self, scope, receive, send):
        assert scope["type"] == WEBSOCKET_TYPE
        self.receive = receive
        self._send = send

    async def send_text(self, data: str) -> None:
        await self._send({"type": WEBSOCKET_SEND_TYPE, "text": data})

    async def send_json(self, data: dict) -> None:
        text_message = json.dumps(data, separators=(",", ":"))
        await self._send({"type": WEBSOCKET_SEND_TYPE, "text": text_message})

    async def send_binary(self, data: bytes) -> None:
        await self._send({"type": WEBSOCKET_SEND_TYPE, "bytes": data})

    async def on_open(self):
        await self._send({"type": WEBSOCKET_ACCEPT_TYPE})

    async def on_close(self):
        try:
            await self._send({"type": WEBSOCKET_CLOSE_TYPE})
        except Exception as e:
            log.error(f"Error sending close message: {e}")



from enum import Enum


class HttpMethod(Enum):
    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PUT = "PUT"
    PATCH = "PATCH"
    OPTIONS = "OPTIONS"


SQLALCHEMY_TYPE_MAPPING = {
    "string": "String",
    "boolean": "Boolean",
    "enum": "Enum",
    "integer": "Integer",
    "numeric": "Numeric",
    "float": "Float",
    "date": "Date",
    "datetime": "DateTime",
    "time": "Time",
    "interval": "Interval",
    "text": "Text",
    "unicode": "Unicode",
    "unicodetext": "UnicodeText",
    "largebinary": "LargeBinary",
    "pickletype": "PickleType",
    "json": "JSON",
}



import click

from inspira.cli.create_app import generate_project
from inspira.cli.create_controller import create_controller_file, create_src_directory
from inspira.cli.generate_database_file import create_database_file
from inspira.cli.generate_model_file import database_file_exists, generate_model_file
from inspira.cli.generate_repository_file import generate_repository_file
from inspira.cli.generate_service_file import generate_service_file
from inspira.migrations.migrations import create_migrations, run_migrations
from inspira.migrations.utils import get_all_module_names

DATABASE_TYPES = ["postgres", "mysql", "sqlite", "mssql"]


@click.group()
def cli():
    pass


@cli.group()
def new():
    pass


@new.command()
@click.argument("name")
@click.option("--only-controller", "only_controller", is_flag=True, required=False)
@click.option("--websocket", "is_websocket", is_flag=True, required=False)
def module(name, only_controller, is_websocket):
    if not name:
        click.echo("Please provide a name for the module")
        return

    if not only_controller and not database_file_exists():
        click.echo(
            "Database file doesn't exist. Please generate one before generating modules"
        )
        return

    create_module_files(name, only_controller, is_websocket)


def create_module_files(name, only_controller, is_websocket):
    create_src_directory()
    create_controller_file(name, is_websocket)

    if not only_controller:
        generate_model_file(name)
        generate_service_file(name)
        generate_repository_file(name)


@new.command()
@click.option("--name", required=True, help="Name of the database.")
@click.option("--type", required=True, help="Database type")
def database(name, type):
    create_database_file(name, type)


@cli.command()
@click.argument("module_name")
@click.option(
    "--empty", nargs=1, type=str, required=False, help="Generate empty migration file."
)
def createmigrations(module_name, empty):
    migration_name = None

    if empty:
        migration_name = empty

    create_migrations(module_name, migration_name)


@cli.command()
@click.argument("module_name", required=False)
def migrate(module_name):
    if module_name:
        module_names = [module_name]
    else:
        module_names = get_all_module_names()

    for module_name in module_names:
        run_migrations(module_name)


@cli.command()
def init():
    generate_project()
    click.echo("App file created successfully.")


if __name__ == "__main__":
    cli()



import re


def convert_to_snake_case(input_string: str) -> str:
    snake_case_string = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", input_string)
    snake_case_string = snake_case_string.replace("-", "_")
    snake_case_string = snake_case_string.lower()

    return snake_case_string


def convert_to_camel_case(input_string: str) -> str:
    space_separated = input_string.replace("_", " ").replace("-", " ")
    capitalized_words = "".join(word.capitalize() for word in space_separated.split())
    camel_case_string = capitalized_words[0] + capitalized_words[1:]

    return camel_case_string



import datetime
import json

from http.cookies import SimpleCookie

from itsdangerous import URLSafeTimedSerializer

from inspira.globals import get_global_app
from inspira.logging import log


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return super().default(obj)


def encode_session_data(session_data, secret_key):
    expiration_time = datetime.datetime.utcnow() + datetime.timedelta(
        seconds=get_global_app().config["SESSION_MAX_AGE"]
    )

    serializer = URLSafeTimedSerializer(secret_key)

    payload = {**session_data, "expiration_time": expiration_time}

    json_session_data = json.dumps(payload, cls=DateTimeEncoder)

    session_token = serializer.dumps(json_session_data)

    return session_token


def decode_session_data(session_token, secret_key):
    try:
        serializer = URLSafeTimedSerializer(secret_key)
        json_session_data = serializer.loads(session_token)
        decoded_payload = json.loads(json_session_data)
        return decoded_payload
    except Exception as e:
        log.error(f"Error decoding session: {e}")
        return None


def get_or_create_session(request):
    session_cookie = get_session_token_from_request(
        request, get_global_app().config["SESSION_COOKIE_NAME"]
    )
    secret_key = get_global_app().secret_key

    if session_cookie:
        try:
            session_data = decode_session_data(session_cookie, secret_key)
            return session_data
        except ValueError:
            log.error("Invalid signature when decoding session")
            raise ValueError("Invalid signature")
    else:
        log.error("No session in cookies")


def get_session_token_from_request(request, session_cookie_name):
    cookies = SimpleCookie(request.get_headers().get("cookie", ""))
    token_cookie = cookies.get(session_cookie_name)
    return token_cookie.value if token_cookie else None



global_app_instance = None


def set_global_app(app, secret_key):
    global global_app_instance
    global_app_instance = app
    global_app_instance.secret_key = secret_key


def get_global_app():
    return global_app_instance



from functools import wraps

from inspira.auth.auth_utils import decode_auth_token
from inspira.requests import RequestContext
from inspira.responses import HttpResponse


def login_required(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        request = RequestContext.get_request()

        token = request.get_session("token")

        if not token or decode_auth_token(token) is None:
            return HttpResponse("Unauthorized", status_code=401)

        return await func(*args, **kwargs)

    return wrapper



import importlib
import inspect
import os
import re
import sys
from typing import Any, Callable, Dict, List

from inspira.config import Config
from inspira.constants import SRC_DIRECTORY
from inspira.enums import HttpMethod
from inspira.globals import set_global_app
from inspira.helpers.error_handlers import (
    default_error_handler,
    handle_method_not_allowed,
    handle_not_found,
)
from inspira.helpers.static_file_handler import handle_static_files
from inspira.logging import log
from inspira.requests import Request, RequestContext
from inspira.utils.session_utils import get_or_create_session
from inspira.utils.controller_parser import parse_controller_decorators
from inspira.utils.dependency_resolver import resolve_dependencies_automatic
from inspira.utils.handler_invoker import invoke_handler
from inspira.websockets import handle_websocket


class Inspira:
    def __init__(self, secret_key=None, config=None):
        self.config = config if config is not None else Config()
        self.secret_key = (
            secret_key if secret_key is not None else self.config["SECRET_KEY"]
        )
        set_global_app(self, self.secret_key)

        self.routes: Dict[str, Dict[str, Callable]] = {
            method.value: {} for method in HttpMethod
        }
        self.error_handler = default_error_handler
        self.middleware: List[Callable] = []
        self.discover_controllers()

    def add_middleware(self, middleware: Callable) -> Callable:
        self.middleware.append(middleware)
        return middleware

    def add_route(self, path: str, method: HttpMethod, handler: Callable) -> None:
        if path in self.routes[method.value]:
            raise AssertionError(
                f"Route with method '{method}' and path '{path}' already exists"
            )

        self.routes[method.value][path] = handler

    def discover_controllers(self) -> None:
        current_dir = os.getcwd()
        src_dir = os.path.join(current_dir, SRC_DIRECTORY)

        for root, _, files in os.walk(src_dir):
            for file_name in files:
                file_path = os.path.join(root, file_name)
                if self._is_controller_file(file_path):
                    rel_path = os.path.relpath(file_path, src_dir)[:-3].replace(
                        os.sep, "."
                    )
                    module_path = f"src.{rel_path}"
                    self._add_routes(module_path)

    def _add_routes(self, file_path: str) -> None:
        try:
            module_name = self._file_path_to_module(file_path)
            src_directory = os.path.abspath(
                os.path.join(file_path, os.pardir, os.pardir)
            )
            sys.path.insert(0, src_directory)
            module = importlib.import_module(module_name)
            for name, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and hasattr(obj, "__path__"):
                    self._add_class_routes(obj)
        except ImportError as e:
            log.error(f"Error importing module {file_path}: {e}")
        finally:
            # Remove the 'src' directory from the Python path after importing
            sys.path.pop(0)

    def _add_class_routes(self, cls) -> None:
        if not hasattr(cls, "__path__"):
            return

        dependencies = resolve_dependencies_automatic(cls)
        instance = cls(*dependencies) if dependencies is not None else cls()

        path_prefix = getattr(cls, "__path__", "")

        for name, method in inspect.getmembers(instance, inspect.ismethod):
            if (
                hasattr(method, "__is_handler__")
                and hasattr(method, "__method__")
                and hasattr(method, "__path__")
            ):
                http_method = getattr(method, "__method__")
                route = getattr(method, "__path__")
                full_route = path_prefix + route
                self.add_route(full_route, http_method, method)

    def _file_path_to_module(self, file_path: str) -> str:
        rel_path = os.path.relpath(file_path, os.getcwd())
        return rel_path.replace(os.sep, ".")

    def _is_controller_file(self, file_path: str) -> bool:
        return file_path.endswith("_controller.py") and parse_controller_decorators(
            file_path
        )

    async def __call__(
        self, scope: Dict[str, Any], receive: Callable, send: Callable
    ) -> None:
        if scope["type"] == "websocket":
            await handle_websocket(scope, receive, send)
        elif scope["type"] == "http":
            await self.process_middlewares(scope, receive, send, self.handle_http)

    async def handle_http(
        self, scope: Dict[str, Any], receive: Callable, send: Callable
    ) -> None:
        request = RequestContext.get_request()
        await self.set_request_session(request)

        method = scope["method"]
        path = scope["path"]

        if path.startswith("/static"):
            await handle_static_files(scope, receive, send, request)
        elif path in self.routes[method]:
            await self.handle_route(method, path, receive, request, scope, send)
        else:
            # Check if the route is present but with a different method
            if any(path in methods for methods in self.routes.values()):
                await handle_method_not_allowed(scope, receive, send)
            else:
                await self.handle_dynamic_route(
                    method, path, request, scope, receive, send
                )

    async def handle_dynamic_route(
        self,
        method: str,
        path: str,
        request: Request,
        scope: Dict[str, Any],
        receive: Callable,
        send: Callable,
    ):
        for route_path, handler in self.routes[method].items():
            if "{" in route_path and "}" in route_path:
                route_pattern = route_path.replace("{", "(?P<").replace("}", ">[^/]+)")
                match = re.fullmatch(route_pattern, path)
                if match:
                    try:
                        params = match.groupdict()
                        response = await invoke_handler(handler, request, scope, params)
                        await response(scope, receive, send)
                        return
                    except Exception as exc:
                        error_response = await self.error_handler(exc)
                        await error_response(scope, receive, send)
        # If no matching route is found, return a 404 response
        await handle_not_found(scope, receive, send)

    async def handle_route(
        self,
        method: str,
        path: str,
        receive: Callable,
        request: Request,
        scope: Dict[str, Any],
        send: Callable,
    ):
        try:
            handler = self.get_handler(method, path)
            response = await self.invoke_handler(handler, request, scope)
            await response(scope, receive, send)
        except Exception as exc:
            await self.handle_error(exc, scope, receive, send)

    def get_handler(self, method: str, path: str):
        return self.routes[method][path]

    async def invoke_handler(self, handler, request: Request, scope: Dict[str, Any]):
        return await invoke_handler(handler, request, scope)

    async def handle_error(self, exc, scope, receive, send):
        error_response = await self.error_handler(exc)
        await error_response(scope, receive, send)

    async def process_middlewares(
        self, scope: Dict[str, Any], receive: Callable, send: Callable, handler
    ):
        request = await self.create_request(receive, scope, send)
        RequestContext.set_request(request)

        for middleware in reversed(self.middleware):
            handler = await middleware(handler)
        response = await handler(scope, receive, send)
        return response

    async def create_request(
        self, receive: Callable, scope: Dict[str, Any], send: Callable
    ) -> Request:
        return Request(scope, receive, send)

    async def set_request_session(self, request: Request) -> None:
        session = get_or_create_session(request)
        request.session = session



import os

import click

from inspira.constants import SRC_DIRECTORY
from inspira.utils import pluralize_word, singularize


def generate_model_file(module_name):
    model_directory = os.path.join(SRC_DIRECTORY, module_name.lower())

    template_path = os.path.join(
        os.path.dirname(__file__), "templates", "model_template.txt"
    )
    repository_file_path = os.path.join(
        model_directory, f"{singularize(module_name.lower())}.py"
    )

    with open(template_path, "r") as template_file, open(
        repository_file_path, "w"
    ) as output_file:
        content = (
            template_file.read()
            .replace(
                "{{module_name_capitalize}}", singularize(module_name.capitalize())
            )
            .replace("{{module_name_plural}}", pluralize_word(module_name))
        )
        output_file.write(content)


def database_file_exists() -> bool:
    main_script_path = "database.py"

    # Check if the main script exists
    if not os.path.isfile(main_script_path):
        click.echo("Main script (database.py) not found.")
        return False

    return True
